#include "ribus_objects.h"
#include "ribus_resources.h"
#include "built_in.h"
#include "FT800_Types.h"


// TFT module connections
sbit FT800_RST at GPIOC_ODR.B2;
sbit FT800_CS at GPIOD_ODR.B13;
sbit FT800_RST_Direction at GPIOC_ODR.B2;
sbit FT800_CS_Direction at GPIOD_ODR.B13;
// End TFT module connections

// Object type constants
// Usage: VTFT stack internally
const VTFT_OT_EVETEXT   = 40;
const VTFT_OT_EVENUMBER = 42;
// ~Object type constants

// Event type constants
// Usage: OnEvent
const VTFT_EVT_UP    = 0;
const VTFT_EVT_DOWN  = 1;
const VTFT_EVT_CLICK = 2;
const VTFT_EVT_PRESS = 3;
// ~Event type constants

// Sound action constants
// Usage: sound event action property and ProcessEvent
const VTFT_SNDACT_NONE  = 0;
const VTFT_SNDACT_PLAY  = 1;
const VTFT_SNDACT_STOP  = 2;
// ~Sound action constants

// Resource loading constants.
// Usage: DrawScreenO and LoadCurrentScreenResToGRAM
const VTFT_LOAD_RES_NONE    = 0x00; // do not load g-ram resources
const VTFT_LOAD_RES_STATIC  = 0x01; // load g-ram resources for static objects
const VTFT_LOAD_RES_DYNAMIC = 0x02; // load g-ram resources for dynamic objects
const VTFT_LOAD_RES_ALL     = 0x03; // load g-ram resources for all objects
// ~Resource loading constants

// Display effect constants
// Usage: DrawScreenO
const VTFT_DISPLAY_EFF_NONE         = 0x00; // no effect when switching between screens
const VTFT_DISPLAY_EFF_LIGHTS_FADE  = 0x04; // backlight: fade out before, fade in after drawing new screen
const VTFT_DISPLAY_EFF_LIGHTS_OFF   = 0x08; // backlight: turn off before, turn on after drawing new screen
// ~Display effect constants

// Stack flags
// Usage: internally used by VTFT stack
const VTFT_INT_REPAINT_ON_DOWN     = 1 << 0;
const VTFT_INT_REPAINT_ON_UP       = 1 << 1;
const VTFT_INT_BRING_TO_FRONT      = 1 << 2;
const VTFT_INT_INTRINSIC_CLICK_EFF = 1 << 3;
// ~Stack flags

// Table of object draw handlers
// Use object type constants to access coresponding object draw handler
const TPointer DrawHandlerTable[44] = {
  0,               // Button draw handler not used
  0,               // CButton draw handler not used
  0,               // ButtonRound draw handler not used
  0,               // CButtonRound draw handler not used
  0,               // Label draw handler not used
  0,               // CLabel draw handler not used
  0,               // Image draw handler not used
  0,               // CImage draw handler not used
  0,               // Circle draw handler not used
  0,               // CCircle draw handler not used
  0,               // CircleButton draw handler not used
  0,               // CCircleButton draw handler not used
  0,               // Box draw handler not used
  0,               // CBox draw handler not used
  0,               // BoxRound draw handler not used
  0,               // CBoxRound draw handler not used
  0,               // Line draw handler not used
  0,               // CLine draw handler not used
  0,               // Polygon draw handler not used
  0,               // CPolygon draw handler not used
  0,               // CheckBox draw handler not used
  0,               // RadioButton draw handler not used
  0,               // ProgressBar draw handler not used
  0,               // Audio draw handler not used
  0,               // EveClock draw handler not used
  0,               // EveGauge draw handler not used
  0,               // EveDial draw handler not used
  0,               // EveKeys draw handler not used
  0,               // CEveKeys draw handler not used
  0,               // EveProgressBar draw handler not used
  0,               // EveScrollBar draw handler not used
  0,               // EveToggle draw handler not used
  0,               // EveSlider draw handler not used
  0,               // EveSpinner draw handler not used
  0,               // EveScreenSaver draw handler not used
  0,               // EveSketch draw handler not used
  0,               // EveButton draw handler not used
  0,               // CEveButton draw handler not used
  0,               // EveGradient draw handler not used
  0,               // CEveGradient draw handler not used
  &DrawEveText,    // EveText draw handler
  0,               // CEveText draw handler not used
  &DrawEveNumber,  // EveNumber draw handler
  0                // CEveNumber draw handler not used
};
// ~Table of draw handler pointers


// Default configuration parameters
const TFT800Display VTFT_FT800_CONFIG_DISPLAY =
{
  48000000,        // Frequency          = main clock frequency
  0,               // OutRenderMode      = 0 normal, 1 write, 2 read
  0,               // RenderReadScanLine = scanline for read render mode
  0,               // RenderWriteTrigger = trigger for write render mode (read only)
  525,             // hCycle             = number if horizontal cycles for display
  43,              // hOffset            = horizontal offset from starting signal
  480,             // hSize              = width resolution
  0,               // hSync0             = hsync falls
  41,              // hSync1             = hsync rise
  286,             // vCycle             = number of vertical cycles for display
  12,              // vOffset            = vertical offset from start signal
  272,             // vSize              = height resolution
  0,               // vSync0             = vsync falls
  10,              // vSync1             = vsync rise
  0,               // Rotate             = rotate display
  0x01B6,          // OutBits            = output bits resolution
  0,               // OutDither          = output number of bits
  0x0000,          // OutSwizzle         = output swizzle
  0,               // OutCSpread         = output clock spread enable
  1,               // PClockPolarity     = clock polarity: 0 - rising edge, 1 - falling edge
  5,               // PClock             = clock prescaler of FT800: - 0 means disable and >0 means 48MHz/pclock
};

const TFT800Touch VTFT_FT800_CONFIG_TOUCH =
{
  3,               // TouchMode        = touch screen mode (2 bits): 0 - off, 1 - oneshot, 2 - frame, 3 - continuous
  1,               // TouchADCMode     = touch screen adc mode (1 bit): 0 - singleended, 1 - differential
  6000,            // TouchCharge      = Touchscreen charge time, units of 6 clocks
  3,               // TouchSettle      = touch screen settle time - 4 bits
  7,               // TouchOversample  = touch screen oversample - 4 bits
  2000,            // TouchRZThreshold = Touchscreen resistance threshold
};

const TFT800TouchTransform VTFT_FT800_CONFIG_TOUCHTRANSFORM =
{
  0x00007EFA,      // TransformA = touch matrix transformation coefficient A
  0xFFFFFF9D,      // TransformB = touch matrix transformation coefficient B
  0xFFF2C8D8,      // TransformC = touch matrix transformation coefficient C
  0xFFFFFFA8,      // TransformD = touch matrix transformation coefficient D
  0xFFFFB43E,      // TransformE = touch matrix transformation coefficient E
  0x011F3BfB,      // TransformF = touch matrix transformation coefficient F
};

const TFT800GPIO VTFT_FT800_CONFIG_GPIO =
{
  0xFC,            // GPIODIR = GPIO direction: 1 - output, 0 - input (8bit wide)
  0xFF,            // GPIO    = GPIO data latch
};

const TFT800PWM VTFT_FT800_CONFIG_PWM =
{
  250,             // Freq = PWM frequency - 14 bits
  64,              // Duty = PWM duty cycle, 0 to 128 is the range
};

const TFT800Interrupt VTFT_FT800_CONFIG_INTERRUPT =
{
  0,               // Flags  = interrupt flags (read only)
  0,               // Enable = global interrupt enable: 1 - enabled, 0 - disabled
  255,             // Mask   = interrupt mask value (individual interrupt enable): 1 - masked/disabled, 0 - enabled
};

const TFT800Sound VTFT_FT800_CONFIG_SOUND =
{
  0,               // Volume
  {0,              // Effect
  0},              // Pitch
  0,               // Play
};

const TFT800Audio VTFT_FT800_CONFIG_AUDIO =
{
  0,               // Volume
  0,               // StartAddress
  0,               // Length
  0,               // ReadPtr
  8000,            // Frequency
  0,               // Format
  0,               // Loop = audio playback mode
  0,               // Play
};

// Global variables

TTouchStat TouchS = {0};


/////////////////////////
TScreen *CurrentScreen = 0;


TScreen Screen1;

TEveText   EveText1;
char       EveText1_Caption[8] = "COUNTER";
TEveNumber EveNumber1;

TEveText *const code Screen1_EveTexts[1] = {
  &EveText1             
};

TEveNumber *const code Screen1_EveNumbers[1] = {
  &EveNumber1           
};


static char IsInsideObject(TObjInfo *AObjInfo, unsigned int X, unsigned int Y) {
  TRect *ptrPressRect = 0;
  TRect *ptrPressCircle = 0;

  if ((AObjInfo->HitX == X) && (AObjInfo->HitY == Y))
    return 1;

  switch (AObjInfo->Type) {
    // EveText
    case VTFT_OT_EVETEXT: {
      ptrPressRect = (TRect *)&(((TEveText *)AObjInfo->Obj)->Left);
      break;
    }
    // EveNumber
    case VTFT_OT_EVENUMBER: {
      ptrPressRect = (TRect *)&(((TEveNumber *)AObjInfo->Obj)->Left);
      break;
    }
  }

  if (ptrPressRect) {
    if ((ptrPressRect->Left <= X) && (ptrPressRect->Left+ptrPressRect->Width-1 >= X) &&
        (ptrPressRect->Top  <= Y) && (ptrPressRect->Top+ptrPressRect->Height-1 >= Y))
      return 1;
    }
  else if (ptrPressCircle) {
    if ((ptrPressCircle->Left <= X) && (ptrPressCircle->Left+ptrPressCircle->Width*2-1 >= X) &&
        (ptrPressCircle->Top  <= Y) && (ptrPressCircle->Top+ptrPressCircle->Width*2-1 >= Y))
      return 1;
    }

  return 0;
}

void DrawEveText(TEveText *AEveText) {
  unsigned long drawOptions;
  
  if (AEveText->Visible) {
    if (AEveText->FontHandle >= 16)
      FT800_Canvas_FontSystem(AEveText->FontHandle, AEveText->Font_Color, AEveText->Opacity);
    else
      FT800_Canvas_Font(AEveText->FontHandle, AEveText->FontName, AEveText->Source, AEveText->Font_Color, AEveText->Opacity);

    if (AEveText->Tag)
      FT800_Canvas_Tag(AEveText->Tag);

    drawOptions = 0;
    if (AEveText->TextAlign == taCenter)
      drawOptions |= _FT800_CP_DRAW_OPT_CENTER;
    else if (AEveText->TextAlign == taCenterX)
      drawOptions |= _FT800_CP_DRAW_OPT_CENTERX;
    else if (AEveText->TextAlign == taCenterY)
      drawOptions |= _FT800_CP_DRAW_OPT_CENTERY;
    else if (AEveText->TextAlign == taRightX)
      drawOptions |= _FT800_CP_DRAW_OPT_RIGHTX;
    FT800_Canvas_CPGraphics_DrawingOptions(drawOptions);

    FT800_Screen_TextCP(AEveText->Left, AEveText->Top, AEveText->Caption);
  }
}

void DrawEveNumber(TEveNumber *AEveNumber) {
  unsigned long drawOptions;
  
  if (AEveNumber->Visible) {
    if (AEveNumber->FontHandle >= 16)
      FT800_Canvas_FontSystem(AEveNumber->FontHandle, AEveNumber->Font_Color, AEveNumber->Opacity);
    else
      FT800_Canvas_Font(AEveNumber->FontHandle, AEveNumber->FontName, AEveNumber->Source, AEveNumber->Font_Color, AEveNumber->Opacity);

    if (AEveNumber->Tag)
      FT800_Canvas_Tag(AEveNumber->Tag);

    drawOptions = 0;
    if (AEveNumber->TextAlign == taCenter)
      drawOptions |= _FT800_CP_DRAW_OPT_CENTER;
    else if (AEveNumber->TextAlign == taCenterX)
      drawOptions |= _FT800_CP_DRAW_OPT_CENTERX;
    else if (AEveNumber->TextAlign == taCenterY)
      drawOptions |= _FT800_CP_DRAW_OPT_CENTERY;
    else if (AEveNumber->TextAlign == taRightX)
      drawOptions |= _FT800_CP_DRAW_OPT_RIGHTX;
    if (AEveNumber->Signed)
      drawOptions |= _FT800_CP_DRAW_OPT_SIGNED;
    drawOptions |= AEveNumber->Text_Length;
    FT800_Canvas_CPGraphics_DrawingOptions(drawOptions);

    FT800_Screen_NumberCP(AEveNumber->Left, AEveNumber->Top, AEveNumber->Value);
  }
}

void SetIdenticalFontSources(const code char *AFontName, unsigned long ASource) {
  char i;
  TEveText   *pEveText;
  TEveNumber *pEveNumber;
  void *const code *ptrO;

  // EveText
  i    = CurrentScreen->EveTextsCount;
  ptrO = CurrentScreen->EveTexts;
  while (i--) {
    pEveText = (TEveText *)(*ptrO);
    if (pEveText->FontName == AFontName)
      pEveText->Source = ASource;
    ptrO++;
  }

  // EveNumber
  i    = CurrentScreen->EveNumbersCount;
  ptrO = CurrentScreen->EveNumbers;
  while (i--) {
    pEveNumber = (TEveNumber *)(*ptrO);
    if (pEveNumber->FontName == AFontName)
      pEveNumber->Source = ASource;
    ptrO++;
  }
}

static void ClearDynObjSource() {
  char i;
  TEveText   *pEveText;
  TEveNumber *pEveNumber;
  void *const code *ptrO;

  // EveText
  i    = CurrentScreen->EveTextsCount;
  ptrO = CurrentScreen->EveTexts;
  while (i--) {
    pEveText = (TEveText *)(*ptrO);
    pEveText->Source = 0xFFFFFFFF;
    ptrO++;
  }

  // EveNumber
  i    = CurrentScreen->EveNumbersCount;
  ptrO = CurrentScreen->EveNumbers;
  while (i--) {
    pEveNumber = (TEveNumber *)(*ptrO);
    pEveNumber->Source = 0xFFFFFFFF;
    ptrO++;
  }
}

void LoadCurrentScreenResToGRAM(char loadOptions) {
  char i;
  long currSource = -1;
  long tmpDynResStart;
  TEveText   *pEveText;
  TEveNumber *pEveNumber;
  void *const code *ptrO;

  // clear dynamic resource addresses first, if necessary
  if (loadOptions & VTFT_LOAD_RES_DYNAMIC)
    ClearDynObjSource();

  // dynamic resources allocation
  if (loadOptions & VTFT_LOAD_RES_DYNAMIC) {

    tmpDynResStart = CurrentScreen->DynResStart;
    if (FT800_RES_BeginLoad(tmpDynResStart)) {
      return;
    }

    // EveText
    i    = CurrentScreen->EveTextsCount;
    ptrO = CurrentScreen->EveTexts;
    while (i--) {
      pEveText = (TEveText *)(*ptrO);
      if (pEveText->FontHandle < 16)
        if (pEveText->Source == 0xFFFFFFFF) {
          currSource = FT800_RES_LoadFont(pEveText->FontName);
          pEveText->Source = currSource;
          SetIdenticalFontSources(pEveText->FontName, currSource);
        }
      ptrO++;
    }

    // EveNumber
    i    = CurrentScreen->EveNumbersCount;
    ptrO = CurrentScreen->EveNumbers;
    while (i--) {
      pEveNumber = (TEveNumber *)(*ptrO);
      if (pEveNumber->FontHandle < 16)
        if (pEveNumber->Source == 0xFFFFFFFF) {
          currSource = FT800_RES_LoadFont(pEveNumber->FontName);
          pEveNumber->Source = currSource;
          SetIdenticalFontSources(pEveNumber->FontName, currSource);
        }
      ptrO++;
    }

    FT800_RES_EndLoad();
  }
}

void DrawObject(TPointer aObj, char aObjType) {
  TDrawHandler drawHandler;

  drawHandler = DrawHandlerTable[aObjType];
  if (drawHandler)
    drawHandler(aObj);
}

void DrawScreenO(TScreen *aScreen, char aOptions) {
  unsigned short cOrder, saveOrder;
  signed   int   actObjOrder;
  unsigned short pwmDuty;
  // counter variables
  char cntEveText;
  char cntEveNumber;
  // pointer variables
  TEveText   *const code *pEveText;
  TEveNumber *const code *pEveNumber;

  // process screen switching effects
  if (aOptions & VTFT_DISPLAY_EFF_LIGHTS_FADE) {
    FT800_PWM_Get(0, &pwmDuty);
    FT800_PWM_FadeOut(pwmDuty, 0, pwmDuty/32? pwmDuty/32 : 1, 1);
  }
  else if (aOptions & VTFT_DISPLAY_EFF_LIGHTS_OFF) {
    FT800_PWM_Get(0, &pwmDuty);
    FT800_PWM_Duty(0);
  }

  if (CurrentScreen != aScreen) {
  // clear active object when drawing to new screen
    memset(&TouchS.ActObjInfo, 0, sizeof(TObjInfo));
  }

  CurrentScreen = aScreen;

  LoadCurrentScreenResToGRAM(aOptions);

  // init counter variables
  cntEveText   = CurrentScreen->EveTextsCount;
  cntEveNumber = CurrentScreen->EveNumbersCount;
  // init pointer variables
  pEveText   = CurrentScreen->EveTexts;
  pEveNumber = CurrentScreen->EveNumbers;

  FT800_Screen_BeginUpdateCP();
  FT800_Canvas_BrushSingleColor(_FT800_BRUSH_STYLE_SOLID, CurrentScreen->Color, 255);
  FT800_Canvas_Tag(0);
  FT800_Screen_Clear(_FT800_CLEAR_ALL);
  FT800_CP_CmdStop();

  actObjOrder = -1;
  if (TouchS.ActObjInfo.Obj)
    if (TouchS.ActObjInfo.Flags & VTFT_INT_BRING_TO_FRONT)
      actObjOrder = TouchS.ActObjInfo.Order;

  cOrder = 0;
  while (cOrder < CurrentScreen->ObjectsCount) {
    saveOrder = cOrder;
    if (pEveText) {
      while ((*pEveText)->Order == cOrder) {
        if (actObjOrder != cOrder) // draw pressed object at the end
          DrawEveText(*pEveText);
        cOrder++;
        pEveText++;
        cntEveText--;
        if (!cntEveText) {
          pEveText = 0;
          break;
        }
      }
      if (saveOrder != cOrder)
        continue;
    }

    if (pEveNumber) {
      while ((*pEveNumber)->Order == cOrder) {
        if (actObjOrder != cOrder) // draw pressed object at the end
          DrawEveNumber(*pEveNumber);
        cOrder++;
        pEveNumber++;
        cntEveNumber--;
        if (!cntEveNumber) {
          pEveNumber = 0;
          break;
        }
      }
      if (saveOrder != cOrder)
        continue;
    }

    cOrder++;
  }

  // draw pressed object now
  if (TouchS.ActObjInfo.Obj)
    DrawObject(TouchS.ActObjInfo.Obj, TouchS.ActObjInfo.Type);

  FT800_Screen_EndUpdate();
  FT800_Screen_Show();

  // process screen switching effects
  if (aOptions & VTFT_DISPLAY_EFF_LIGHTS_FADE) {
    FT800_PWM_FadeIn(0, pwmDuty, 1, 3);
  }
  else if (aOptions & VTFT_DISPLAY_EFF_LIGHTS_OFF) {
    FT800_PWM_Duty(pwmDuty);
  }

}

void DrawScreen(TScreen *aScreen) {
  if (aScreen != CurrentScreen)
    DrawScreenO(aScreen, VTFT_LOAD_RES_ALL | VTFT_DISPLAY_EFF_LIGHTS_FADE);
  else
    DrawScreenO(aScreen, VTFT_LOAD_RES_NONE);
}

char GetActiveObjectByXY(int X, int Y, TObjInfo *AObjInfo) {
  char i;
  int  hiOrder;
  TEveText   *pEveText;
  TEveNumber *pEveNumber;
  void *const code *ptrO;

  // clear current object info
  memset(AObjInfo, 0, sizeof(TObjInfo));

  // Find object with highest order at specified position.
  // Objects lists are sorted by object order ascending.
  hiOrder = -1;

  // EveText
  i    = CurrentScreen->EveTextsCount;
  ptrO = CurrentScreen->EveTexts+CurrentScreen->EveTextsCount-1;
  while (i--) {
    pEveText = (TEveText *)(*ptrO);
    if (pEveText->Order < hiOrder)
      break;
    if (pEveText->Active) {
      if ((pEveText->Left <= X) && (pEveText->Left+pEveText->Width-1 >= X) &&
          (pEveText->Top  <= Y) && (pEveText->Top+pEveText->Height-1 >= Y)) {
        AObjInfo->Obj   = (void *)pEveText;
        AObjInfo->Type  = VTFT_OT_EVETEXT;
        AObjInfo->Order = pEveText->Order;
        AObjInfo->Flags = VTFT_INT_BRING_TO_FRONT;

        hiOrder         = pEveText->Order;

        break;
      }
    }
    ptrO--;
  }

  // EveNumber
  i    = CurrentScreen->EveNumbersCount;
  ptrO = CurrentScreen->EveNumbers+CurrentScreen->EveNumbersCount-1;
  while (i--) {
    pEveNumber = (TEveNumber *)(*ptrO);
    if (pEveNumber->Order < hiOrder)
      break;
    if (pEveNumber->Active) {
      if ((pEveNumber->Left <= X) && (pEveNumber->Left+pEveNumber->Width-1 >= X) &&
          (pEveNumber->Top  <= Y) && (pEveNumber->Top+pEveNumber->Height-1 >= Y)) {
        AObjInfo->Obj   = (void *)pEveNumber;
        AObjInfo->Type  = VTFT_OT_EVENUMBER;
        AObjInfo->Order = pEveNumber->Order;
        AObjInfo->Flags = VTFT_INT_BRING_TO_FRONT;

        hiOrder         = pEveNumber->Order;

        break;
      }
    }
    ptrO--;
  }

  if (AObjInfo->Obj) {
    AObjInfo->HitX = X;
    AObjInfo->HitY = Y;
    return 1;
  }
  else {
    return 0;
  }
}

char GetActiveObjectByTag(char ATag, TObjInfo *AObjInfo) {
  char i;
  TEveText   *pEveText;
  TEveNumber *pEveNumber;
  void *const code *ptrO;

  // clear current object info
  memset(AObjInfo, 0, sizeof(TObjInfo));

  // Find object with specified tag value.

  // EveText
  i    = CurrentScreen->EveTextsCount;
  ptrO = CurrentScreen->EveTexts+CurrentScreen->EveTextsCount-1;
  while (i--) {
    pEveText = (TEveText *)(*ptrO);
    if (pEveText->Tag == ATag) {
      if (pEveText->Active) {
        AObjInfo->Obj   = (void *)pEveText;
        AObjInfo->Type  = VTFT_OT_EVETEXT;
        AObjInfo->Order = pEveText->Order;
        AObjInfo->Flags = VTFT_INT_BRING_TO_FRONT;
      }
      break;
    }
    ptrO--;
  }

  // EveNumber
  i    = CurrentScreen->EveNumbersCount;
  ptrO = CurrentScreen->EveNumbers+CurrentScreen->EveNumbersCount-1;
  while (i--) {
    pEveNumber = (TEveNumber *)(*ptrO);
    if (pEveNumber->Tag == ATag) {
      if (pEveNumber->Active) {
        AObjInfo->Obj   = (void *)pEveNumber;
        AObjInfo->Type  = VTFT_OT_EVENUMBER;
        AObjInfo->Order = pEveNumber->Order;
        AObjInfo->Flags = VTFT_INT_BRING_TO_FRONT;
      }
      break;
    }
    ptrO--;
  }

  if (AObjInfo->Obj) {
    AObjInfo->HitTag = ATag;
    return 1;
  }
  else {
    return 0;
  }
}

static void ProcessEvent(TEvent *pEvent) {
  if (pEvent) {
    if (pEvent->Sound.SndAct == VTFT_SNDACT_PLAY) 
      FT800_Sound_SetAndPlay(pEvent->Sound.Effect, pEvent->Sound.Pitch, pEvent->Sound.Volume);
    else if (pEvent->Sound.SndAct == VTFT_SNDACT_STOP) 
      FT800_Sound_Stop();
    if (pEvent->Action) 
      pEvent->Action();
  }
}

static void ProcessCEvent(TCEvent *pEventC) {
  if (pEventC) {
    if (pEventC->Sound.SndAct == VTFT_SNDACT_PLAY) 
      FT800_Sound_SetAndPlay(pEventC->Sound.Effect, pEventC->Sound.Pitch, pEventC->Sound.Volume);
    else if (pEventC->Sound.SndAct == VTFT_SNDACT_STOP) 
      FT800_Sound_Stop();
    if (pEventC->Action) 
      pEventC->Action();
  }
}

static void OnEvent(TObjInfo *AObjInfo, char AEventType){
  TEvent **ppEvent;
  TEvent  *pEvent = 0;

  switch (AObjInfo->Type) {
    // EveText
    case VTFT_OT_EVETEXT: {
      ppEvent = &(((TEveText *)AObjInfo->Obj)->OnUp);
      pEvent  = ppEvent[AEventType];
      break;
    }
    // EveNumber
    case VTFT_OT_EVENUMBER: {
      ppEvent = &(((TEveNumber *)AObjInfo->Obj)->OnUp);
      pEvent  = ppEvent[AEventType];
      break;
    }
  } // end switch

  if (pEvent) {
    ProcessEvent(pEvent);
  }
}

static void Process_TP_Press() {
  // Screen Event
  if (CurrentScreen->Active)
    if ((CurrentScreen->SniffObjectEvents) || (!TouchS.ActObjInfo.Obj))
      ProcessEvent(CurrentScreen->OnPress);

  // Object Event
  if (!TouchS.ActObjInfo.Obj)
    return;

  OnEvent(&TouchS.ActObjInfo, VTFT_EVT_PRESS);
}

static void Process_TP_Up() {
  char     isClick;
  TObjInfo actObj;

  // Screen Event
  if (CurrentScreen->Active)
    if ((CurrentScreen->SniffObjectEvents) || (!TouchS.ActObjInfo.Obj))
      ProcessEvent(CurrentScreen->OnUp);

  actObj = TouchS.ActObjInfo;
  // Cler active object info
  memset(&TouchS.ActObjInfo, 0, sizeof(TObjInfo));

  // Object Event
  if (!actObj.Obj)
    return;

  isClick = IsInsideObject(&actObj, TouchS.X, TouchS.Y);

  if (actObj.Flags & VTFT_INT_REPAINT_ON_UP)
    DrawScreen(CurrentScreen);

  OnEvent(&actObj, VTFT_EVT_UP);
  if (isClick)
    OnEvent(&actObj, VTFT_EVT_CLICK);
}

static void Process_TP_Down() {
  // Search for active object
  if (TouchS.Tag) {        // objects must not have zero for tag value
    if (TouchS.Tag != 255) // can not search objects by default tag value
      GetActiveObjectByTag(TouchS.Tag, &TouchS.ActObjInfo);
    if (!TouchS.ActObjInfo.Obj) // object not found by tag, search by coordinates
      GetActiveObjectByXY(TouchS.X, TouchS.Y, &TouchS.ActObjInfo);
  }

  // Screen Event
  if (CurrentScreen->Active)
    if ((CurrentScreen->SniffObjectEvents) || (!TouchS.ActObjInfo.Obj))
      ProcessEvent(CurrentScreen->OnDown);

  // Object Event
  if (!TouchS.ActObjInfo.Obj)
    return;

  if (TouchS.ActObjInfo.Flags & VTFT_INT_REPAINT_ON_DOWN)
    DrawScreen(CurrentScreen);

  OnEvent(&TouchS.ActObjInfo, VTFT_EVT_DOWN);
}

static void Process_TP_TagChange() {
  // Screen Event
  if (CurrentScreen->Active)
    ProcessEvent(CurrentScreen->OnTagChange);
}

void ProcessVTFTStack() {
  char         Tag, oldTag;
  unsigned int X, Y;

  oldTag = TouchS.Tag;

  if (FT800_Touch_GetTagXY(&X, &Y) == 1) {
    FT800_Touch_GetTag(&Tag);

    TouchS.Tag = Tag;
    TouchS.X = X;
    TouchS.Y = Y;

    if (!TouchS.Pressed) {
      TouchS.Pressed = 1;
      Process_TP_Down();
    }

    Process_TP_Press();
  }
  else if (TouchS.Pressed) {
    Process_TP_Up();

    TouchS.Tag = 0;
    TouchS.X   = X;
    TouchS.Y   = Y;

    TouchS.Pressed = 0;
  }

  if (oldTag != TouchS.Tag)
    Process_TP_TagChange();
}

static void InitObjects() {
  // Screen1: Init block start
  Screen1.Color             = 0x3679;
  Screen1.Width             = 480;
  Screen1.Height            = 272;
  Screen1.ObjectsCount      = 2;
  Screen1.EveTextsCount     = 1;
  Screen1.EveTexts          = Screen1_EveTexts;
  Screen1.EveNumbersCount   = 1;
  Screen1.EveNumbers        = Screen1_EveNumbers;
  Screen1.DynResStart       = 0;
  Screen1.Active            = 1;
  Screen1.SniffObjectEvents = 0;
  Screen1.OnUp              = 0;
  Screen1.OnDown            = 0;
  Screen1.OnTagChange       = 0;
  Screen1.OnPress           = 0;

  EveText1.OwnerScreen = &Screen1;
  EveText1.Order       = 0;
  EveText1.Visible     = 1;
  EveText1.Opacity     = 255;
  EveText1.Tag         = 255;
  EveText1.Left        = 116;
  EveText1.Top         = 50;
  EveText1.Width       = 215;
  EveText1.Height      = 53;
  EveText1.Caption     = EveText1_Caption;
  EveText1.TextAlign   = taNone;
  EveText1.FontName    = Tahoma_48x48_BoldItalic;
  EveText1.Font_Color  = 0xFFFF;
  EveText1.FontHandle  = 1;
  EveText1.Source      = -1UL;
  EveText1.Active      = 1;
  EveText1.OnUp        = 0;
  EveText1.OnDown      = 0;
  EveText1.OnClick     = 0;
  EveText1.OnPress     = 0;

  EveNumber1.OwnerScreen = &Screen1;
  EveNumber1.Order       = 1;
  EveNumber1.Visible     = 1;
  EveNumber1.Opacity     = 255;
  EveNumber1.Tag         = 255;
  EveNumber1.Left        = 200;
  EveNumber1.Top         = 140;
  EveNumber1.Width       = 40;
  EveNumber1.Height      = 53;
  EveNumber1.Text_Length = 0;
  EveNumber1.TextAlign   = taNone;
  EveNumber1.FontName    = Tahoma_48x48_BoldItalic;
  EveNumber1.Font_Color  = 0xFFFF;
  EveNumber1.FontHandle  = 1;
  EveNumber1.Source      = -1UL;
  EveNumber1.Value       = 0;
  EveNumber1.Signed      = 0;
  EveNumber1.Active      = 1;
  EveNumber1.OnUp        = 0;
  EveNumber1.OnDown      = 0;
  EveNumber1.OnClick     = 0;
  EveNumber1.OnPress     = 0;

}

void Init_MCU() {
  // Setup FT800 SPI interface
  SPI1_Init_Advanced(_SPI_FPCLK_DIV8, _SPI_MASTER | _SPI_8_BIT |
  		   _SPI_CLK_IDLE_LOW | _SPI_FIRST_CLK_EDGE_TRANSITION |
  		   _SPI_MSB_FIRST | _SPI_SS_DISABLE | _SPI_SSM_ENABLE | _SPI_SSI_1,
  		   &_GPIO_MODULE_SPI1_PB345);
}

void InitVTFTStack() {
  Init_MCU();

  SPI3_Init();

  // Init all dynamic objects
  InitObjects();

  // Init FT800 controller core and library stack
  FT800_Init();

  FT800_Core_ClockSource(_FT800_CLK_SOURCE_EXTERNAL);
  FT800_Core_ClockPLL(_FT800_CLK_PLL_48MHz);

  // Internal modules setup
  FT800_Display_SetConfig(&VTFT_FT800_CONFIG_DISPLAY);

  FT800_Audio_SetConfig(&VTFT_FT800_CONFIG_AUDIO);

  FT800_Sound_SetConfig(&VTFT_FT800_CONFIG_SOUND);

  FT800_Interrupt_SetConfig(&VTFT_FT800_CONFIG_INTERRUPT);

  FT800_PWM_SetConfig(&VTFT_FT800_CONFIG_PWM);

  FT800_GPIO_SetConfig(&VTFT_FT800_CONFIG_GPIO);

  FT800_Touch_SetConfig(&VTFT_FT800_CONFIG_TOUCH);

  FT800_TouchTransform_SetConfig(&VTFT_FT800_CONFIG_TOUCHTRANSFORM);

  // Draw start screen
  DrawScreen(&Screen1);
}